package org.hardknockjava.houseofcards;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Random;

public class DeckOfCards {
    public enum Suit {
        CLUBS, DIAMONDS, HEARTS, SPADES
    }

    public enum Rank {
        DEUCE(2), TREY(3), FOUR(4), FIVE(5), SIX(6), SEVEN(7), EIGHT(8), NINE(9), TEN(10), JACK(11), QUEEN(12),
        KING(13), ACE(14);

        private int value;

        private Rank(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }
    }

    private List<Card> deck = new ArrayList<Card>();

    public static void main(String[] args) {
        // Create the deck.
        DeckOfCards house = new DeckOfCards();
        house.deal("Cards in creation order");

        // Sort the deck into descending order.
        house.sortByValue();

        // Display the first cards in the deck.
        house.deal("Cards in descending (natural) order");

        // Shuffle the cards.
        house.shuffle();

        // Display the first few cards in the shuffled deck.
        house.deal("Cards after shuffle");
    }

    private DeckOfCards() {
        // Create a random number generator. The seed changes
        // with the date and time, so the outcome differs
        // with every new deck.
        Random shuffleOrder = new Random((new Date()).getTime());

        // Create the Cards and save them in deck.
        for (Suit suit : Suit.values()) {
            for (Rank rank : Rank.values()) {
                Card card = new Card(suit, rank, shuffleOrder.nextInt());
                deck.add(card);
            }
        }
    }

    // sortByValue() sorts the cards in descending order,
    // which is the natural ordering of Card.
    private void sortByValue() {
        deck.sort(null);
    }

    // shuffle() sorts the cards by shuffleOrder, using
    // a Comparator because this isn't the natural order.
    private void shuffle() {
        deck.sort(new Comparator<Card>() {
            @Override
            public int compare(Card card1, Card card2) {
                return card1.getShuffleOrder() - card2.getShuffleOrder();
            }
        });
    }

    // Display the first half-dozen cards.
    private void deal(String title) {
        System.out.println("\n\n" + title);
        for (int i = 0; i < 15; i++) {
            Card card = deck.get(i);
            System.out.println("   " + card.rank + " of " + card.suit);
        }
    }

    ///////////////////////////////////

    private static class Card implements Comparable<Card> {
        Suit suit;
        Rank rank;
        int shuffleOrder;

        public Card(Suit suit, Rank rank, int shuffleOrder) {
            super();
            this.suit = suit;
            this.rank = rank;
            this.shuffleOrder = shuffleOrder;
        }

        public Rank getRank() {
            return rank;
        }

        public Suit getSuit() {
            return suit;
        }
        
        public int getShuffleOrder() {
            return shuffleOrder;
        }

        @Override
        // compareTo() sorts the cards in descending order.
        public int compareTo(Card other) {
            int diff = this.getRank().getValue() - other.getRank().getValue();

            // If the ranks are the same, calculate diff as the
            // difference between the values of the suits.
            if (diff == 0) {
                diff = this.getSuit().ordinal() - other.getSuit().ordinal();
            }

            // Since we want to sort in descending order, return
            // the opposite of diff.
            return -1 * diff;
        }
    }
}
